// Copyright (c) 2000-2001 Brad Hughes <bhughes@trolltech.com>
//
// Use, modification and distribution is allowed without limitation,
// warranty, or liability of any kind.
//

#include <QObject>
#include <QApplication>
#include <QtGlobal>
#include <QDir>
#include <QSettings>

#include "outputalsa.h"
#include "constants.h"
#include "buffer.h"
#include "visual.h"

#include <stdio.h>
#include <string.h>

#include <iostream>

//extern Q_EXPORT QApplication* qApp;


void OutputALSA::stop()
{
    user_stop = TRUE;
}

void OutputALSA::status()
{
    long ct = (total_written - latency()) / bps;

    if (ct < 0)
        ct = 0;

    if (ct > current_seconds)
    {
        current_seconds = ct;
        elapsed_seconds = current_seconds;
//         m_stat = OutputStatus (current_seconds, total_written, lr, lf, lp, lc);
//         emit statusChanged();
        dispatch(OutputState (current_seconds, total_written, lr, lf, lp, lc));
        //statusChanged();  
        //OutputEvent e(current_seconds, total_written, lr, lf, lp, lc);
        //dispatch(e);

    }
}

long OutputALSA::written()
{
    return total_written;
}

void OutputALSA::seek(long pos)
{
    recycler()->mutex()->lock()
    ;
    recycler()->clear();
    recycler()->mutex()->unlock();

    total_written = (pos * bps);
    current_seconds = -1;
}

OutputALSA::OutputALSA(QObject * parent, const QString &d)
        : Output(parent), audio_device(d), inited(FALSE), paus(FALSE), play(FALSE),
        user_stop(FALSE),
        total_written(0), current_seconds(-1),
        bps(1), lf(-1), lc(-1), lp(-1),
        do_select(TRUE),
        audio_fd(-1)
{
    QSettings settings(QDir::homePath()+"/.qmmp/qmmprc", QSettings::IniFormat);
    QString dev_name = settings.value("ALSA/device","default").toString();
    pcm_name = strdup(dev_name.toAscii().data());
    stream = SND_PCM_STREAM_PLAYBACK;
    snd_pcm_hw_params_alloca(&hwparams);
    pcm_handle = 0;
}

OutputALSA::~OutputALSA()
{
    if (pcm_handle)
    {
        qDebug("closing pcm_handle");
        snd_pcm_close(pcm_handle);
        qDebug("closing pcm_handle .. 1");
        pcm_handle = 0;
        qDebug("closing pcm_handle .. done");
    }
    dispatch(OutputState (OutputState::Stopped));

}

void OutputALSA::configure(long freq, int chan, int prec, int brate)
{
    qDebug("configure");
    // we need to configure
    if (freq != lf || chan != lc || prec != lp)
    {
        // we have already configured, but are changing settings...
        // reset the device
        //resetDSP();

        lf = freq;
        lc = chan;
        lp = prec;
        //bps = 16*2/8;
        bps = freq * chan * (prec / 8);
        snd_pcm_hw_params_alloca(&hwparams);
        if (snd_pcm_hw_params_any(pcm_handle, hwparams) < 0)
        {
            fprintf(stderr, "Can not configure this PCM device.\n");
            return;
        }

        uint rate = lf; /* Sample rate */
        uint exact_rate = lf;   /* Sample rate returned by */
        /* snd_pcm_hw_params_set_rate_near */
        int dir;          /* exact_rate == rate --> dir = 0 */
        /* exact_rate < rate  --> dir = -1 */
        /* exact_rate > rate  --> dir = 1 */
        int periods = 2;       /* Number of periods */
        snd_pcm_uframes_t periodsize = 8192; /* Periodsize (bytes) */
        /* Set access type. This can be either    */
        /* SND_PCM_ACCESS_RW_INTERLEAVED or       */
        /* SND_PCM_ACCESS_RW_NONINTERLEAVED.      */
        /* There are also access types for MMAPed */
        /* access, but this is beyond the scope   */
        /* of this introduction.                  */
        /* load settings from config */
        QSettings settings(QDir::homePath()+"/.qmmp/qmmprc", QSettings::IniFormat);
        settings.beginGroup("ALSA");
        uint buffer_time = settings.value("buffer_time",500).toUInt()*1000;
        uint period_time = settings.value("period_time",100).toUInt()*1000;
        settings.endGroup();

        if (snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
        {
            qWarning("Error setting access.\n");
            return;
        }


        if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, SND_PCM_FORMAT_S16_LE) < 0)
        {
            qDebug("Error setting format.\n");
            return;
        }


        exact_rate = rate;// = 11000;
        qDebug("OutputALSA: frequency=%d, channels=%d, bitrate=%d",
                                                     rate, chan, brate);
        if (snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, &exact_rate, 0) < 0)
        {
            qWarning("Error setting rate.\n");
            return;
        }
        if (rate != exact_rate)
        {
            qWarning("The rate %d Hz is not supported by your hardware.\n==> Using %d Hz instead.\n", rate, exact_rate);
        }

        uint c = lc;
        if (snd_pcm_hw_params_set_channels_near(pcm_handle, hwparams, &c) < 0)
        {
            qWarning("Error setting channels.\n");
            return;
        }

        /*if (snd_pcm_hw_params_set_periods(pcm_handle, hwparams, periods, 0) < 0)
        {
            fprintf(stderr, "Error setting periods.\n");
            return;
        }*/
        /*snd_pcm_uframes_t frames = snd_pcm_bytes_to_frames(pcm_handle, periodsize * periods);

        if (snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, &frames) < 0)
        {
            qWarning(stderr, "Error setting buffersize.\n");
            return;
        }*/
        if (snd_pcm_hw_params_set_period_time_near(pcm_handle, hwparams,
                            &period_time ,0) < 0 )
        { 
            fprintf(stderr, "Error setting HW buffer.\n");
            return;
        }
        if (snd_pcm_hw_params_set_buffer_time_near(pcm_handle, hwparams,
                            &buffer_time ,0) < 0 )
        { 
            fprintf(stderr, "Error setting HW buffer.\n");
            return;
        }
        if (snd_pcm_hw_params(pcm_handle, hwparams) < 0)
        {
            qWarning("OutputALSA: Error setting HW params.");
            return;
        }
        prepareVisuals();
        qDebug("configure end");
    }
}

void OutputALSA::reset()
{
    qDebug("reset()");
    if (pcm_handle)
    {
        snd_pcm_close(pcm_handle);
        pcm_handle = 0;
    }
    if (snd_pcm_open(&pcm_handle, pcm_name, stream, SND_PCM_NONBLOCK) < 0)
    {
        qWarning ("Error opening PCM device %s\n", pcm_name);
        return;
    }
    qDebug("reset() ends");
}


void OutputALSA::pause()
{
    paus = (paus) ? FALSE : TRUE;
    {
        int state = paus ? OutputState::Paused: OutputState::Playing;
        dispatch(OutputState((OutputState::Type) state));
    }

}

void OutputALSA::post()
{
    /*if (audio_fd < 1)
        return;

    int unused;
    ioctl(audio_fd, SNDCTL_DSP_POST, &unused);*/
}

void OutputALSA::sync()
{
    /*if (audio_fd < 1)
        return;

    int unused;
    ioctl(audio_fd, SNDCTL_DSP_SYNC, &unused);*/
}


void OutputALSA::resetDSP()
{
    /*if (audio_fd < 1)
        return;

    int unused;
    ioctl(audio_fd, SNDCTL_DSP_RESET, &unused);*/
}


bool OutputALSA::initialize()
{
    inited = paus = play = user_stop = FALSE;

    //reset();
    if (!pcm_handle < 0)
        return FALSE;

    current_seconds = -1;
    total_written = 0;
    if (snd_pcm_open(&pcm_handle, pcm_name, stream, SND_PCM_NONBLOCK) < 0)
    {
        qWarning ("Error opening PCM device %s\n", pcm_name);
        return FALSE;
    }

    inited = TRUE;
    return TRUE;
}


long OutputALSA::latency()
{
    ulong used = 0;

    /*if (! paus)
    {
        if (ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &used) == -1)
            used = 0;
    }*/

    return used;
}

void OutputALSA::run()
{

    mutex()->lock();
    if (! inited)
    {
        mutex()->unlock();
        return;
    }

    play = TRUE;

    mutex()->unlock();
    //fd_set afd;
    //struct timeval tv;
    Buffer *b = 0;
    bool done = FALSE;
    unsigned long n = 0;
    long m = 0, f = 0;
    snd_pcm_uframes_t l;
    //FD_ZERO(&afd);
    {
        //m_stat = OutputStatus (OutputStatus::Playing);
        //emit statusChanged();
        dispatch(OutputState (OutputState::Playing));
    }
    while (! done)
    {
        mutex()->lock();
        recycler()->mutex()->lock();

        done = user_stop;

        while (! done && (recycler()->empty() || paus))
        {
            mutex()->unlock();
            recycler()->cond()->wakeOne();
            recycler()->cond()->wait(recycler()->mutex());
            mutex()->lock();
            done = user_stop;
            status();

        }

        if (! b)
        {
            b = recycler()->next();
            if (b->rate)
                lr = b->rate;
        }

        recycler()->cond()->wakeOne();
        recycler()->mutex()->unlock();

        if (b)
        {

            l = qMin(int(2048), int(b->nbytes - n));
            l = snd_pcm_bytes_to_frames(pcm_handle, l);
            //l = qMin(aval, int(l));
            while(l>0)
            {
                m = snd_pcm_writei (pcm_handle, b->data+n, l);

                if (m > 0)
                {
                    n += snd_pcm_frames_to_bytes(pcm_handle, m);
                    l -= m;
                    status();
                    dispatchVisual(b, total_written, lc, lp); 
                    //qDebug ("Played %d bytes \n", m);
                }
                else if (m == -EAGAIN)
                {
                    /*if (snd_pcm_wait(pcm_handle, 1000) < 0)
                        qDebug ("snd_pcm_wait() failed");*/
                    //while(snd_pcm_resume(pcm_handle) == -EAGAIN)
                    usleep(500);
                        
                }
                else if (m == -EPIPE)
                {
                    qDebug ("underrun!");
                    if ((m = snd_pcm_prepare(pcm_handle)) < 0)
                    {
                        qDebug ("Can't recover after underrun: %s",
                                snd_strerror(m));
                        /* TODO: reopen the device */
                        break;
                    }
                }
                else if (m == -ESTRPIPE)
                {
                    qDebug ("Suspend, trying to resume");
                    while ((m = snd_pcm_resume(pcm_handle))
                            == -EAGAIN)
                        sleep (1);
                    if (m < 0)
                    {
                        qDebug ("Failed, restarting");
                        if ((m = snd_pcm_prepare(pcm_handle))
                                < 0)
                        {
                            qDebug ("Failed to restart "
                                    "device: %s.",
                                    snd_strerror(m));
                            break;
                        }
                    }
                }
                else if (m < 0)
                {
                    qDebug ("Can't play: %s", snd_strerror(m));
                    break;
                }
                /*else
                {
                    n += m;
                    l -= m;
                    status();
                    //qDebug ("Played %d bytes \n", m);
                }*/
            }
            status();
            // force buffer change
            
            total_written += n; //snd_pcm_frames_to_bytes(pcm_handle, n);
            n = b->nbytes;
            m = 0;
        }
        if (n == b->nbytes)
        {
            recycler()->mutex()->lock()
            ;
            recycler()->done();
            recycler()->mutex()->unlock();
            b = 0;
            n = 0;
        }
        mutex()->unlock();
    }

    mutex()->lock()
    ;
    //if (! user_stop)
    //sync();
    //resetDSP();

    play = FALSE;

    {
        //m_stat = OutputStatus (OutputStatus::Stopped);
        //emit statusChanged();
        dispatch(OutputState (OutputState::Stopped)); 
        /*stat = OutputEvent::Stopped;
        OutputEvent e((OutputEvent::Type) stat);
        dispatch(e);*/
    }

    mutex()->unlock();

}

